# Purpose:      filter procedure file for ecachepanic module
# Author:       Carsten Lorbeer
# Date:         20/04/2000
#
# SCCS:
# %Z%%M%    %I% %E% %U%


proc pscanfile {} {

# for the first time initialize all parameter

    if { [ ilookup -d "" sdata filename ] == "" } {
    	ddl print debug "First Run : Initialize Parmeters"
	# Read Module Parameter
        set Filename [ getModuleParam filename "" ]
        set Scanlevel [ getModuleParam scanlevel "incr" ]
      
        
        toe_send [ locate filename ] define data 0 "$Filename"
        toe_send [ locate scanlevel ] define data 0 "$Scanlevel"

        # cache Constants locally 
        define sdata filename "$Filename"
        define sdata scanlevel "$Scanlevel"
	
	#Read persistant Data from the disk
	define sdata mtime [ getAttribute mtime 0 ]
        define sdata seek [ getAttribute seek 0 ] 
	

 

    }
    # Request objects from Panic Pattern Table
    set object_pp [ locate panicpattern ]
    set pentryobject [ locate *patternEntry ]
    set cmobject [ locate *currentMatch ]
    set pmobject [ locate *previousMatch ]
    
    # Request objects from Panics per cpu Table
    set object_cp [ locate cpupanic ]
    set cpupanictbl [ locate *cpupanicTable ]
    set cputbl [ locate *cpupanicEntry ]
    set idtable [ locate *Id ]
    set rowstatusid [ locate *cpupanicEntry.rowstatus ]

    # read Pattern and Patternnames from Panic Patterntable
    define sdata plist [ receivePatternList ]
    define sdata pnlist [ receivePatternNameList ]
    define sdata kplist [ receiveKeyPatternList ]
    
    set rtValue [ srs_pscan ]
 
    # store persistent data
    setAttribute seek [ ilookup -d 0 sdata seek ]
    setAttribute mtime [ ilookup -d 0 sdata mtime ]
    
    # if no matches where found, we can save time.
    if { $rtValue == "" } {
    	ddl print debug "srs_pscan returned no matches..."
    	return ""
    }   
    
    
    # There are two cases. Firstly, we allready created a entry for a
    # specific cpu and patter or we have to update an existing one.
    
     foreach { linenumber date ts message name key } $rtValue {
        set index [ toe_send $pentryobject lookup -d 0 indexByName $name ]
    	toe_begin $object_pp
		set plistentry [ lookup -d "" plist $name ]
		if { $plistentry != "" } {

			if { $index != 0 } {
				set cmatch [ toe_send $cmobject lookup -d 0 data $index ]
				toe_send $pmobject define data $index $cmatch
				toe_send $cmobject define data $index $ts
 				define cmlist $name $ts
				define pmlist $name $cmatch
				
			}
			
		} 	    	    	
    	toe_end
     	if { $key != "" } {
		# Make sure, that key is an integer
		# Othervise set $key to -1, which is never a valid CPUID
		if { ![ isInt $key ] } {
			set key -1
		}
		# 
		# Create a unique key 
     		set uniqueId [ format "%s%d" $name $key ]
     		set maxIdx 0
		
      		if { [ toe_send $cputbl exists indexByName $uniqueId ] == 1 } {
     			ddl print debug "Update existing row $uniqueId ..."
			
			# get the data from the table row, we want to modify
    			set entry [ toe_send $cputbl getRowData $uniqueId ]
 
    			lextract $entry 1 rowdata
    			lextract $rowdata 0 rowstatus 1 id 2 cpu 4 previousp 6 currentp 7 newpanic 8 delta 9 type
    			set delta [ toe_send $object_pp sec2day [ expr $ts-$currentp ] ]
    			set cpd [ sec2date $currentp ]
    			set ppd [ sec2date $ts ]
    			
    			set idx [ toe_send $cputbl lookup indexByName $uniqueId ]
    			
			#update row data
    			toe_send $cputbl setValue $idx [ list $id $cpu $cpd $currentp $ppd $ts 1 $delta $type ]
    			toe_send $rowstatusid setRowStatus $idx $id 1
			
    			ddl print debug "Row update was successfull..."
    		} else {
    			ddl print debug "Create new row $uniqueId..."
			
    			set maxIdx [ toe_send $cputbl getTableDepth ]
     			incr maxIdx 
     			set cpd [ sec2date $ts ]
     			toe_send $cputbl setValue $maxIdx [ list $uniqueId $key "" 0 $cpd $ts 1 "" "$name" ]
    			toe_send $rowstatusid setRowStatus $maxIdx $uniqueId 1
			
    			ddl print debug "Row created successfully..."

     		}
		
		# Save persistant Data and refresh Panic per CPU table
     		toe_send $cpupanictbl syncCPUList

    		toe_send $cpupanictbl setTrapInfo refreshOID

    	}    	
    }
    
    # Save persistant Data and display matches
    toe_send $object_pp syncPatternList
    return $rtValue
}

#
# A key may not contain whitespaces
#  
proc nameToKey { name } {
    regsub -all {\.} $name {/} key
    regsub -all { } $key {_} key
    return $key
}


proc nameToKey { name } {
    regsub -all {\.} $name {/} key
    regsub -all { } $key {_} key
    return $key
}
#
# This function is important, to make sure, that key is
# a numeric value
#
proc isInt { number } {
	regexp {^-?[0-9]+$} $number
}
#
# Save Patternlist to disk
#
proc syncPatternList {} {
    ddl print debug "syncPatternList..."
    set patternList ""

    sliceforeach name pattern plist {
 	set desc [ lookup -d "" dlist $name ]
 	set kpattern [ lookup -d "" kplist $name ]
	set rowstatus [ lookup -d "1" rowstatuslist $name ]
	set currentmatch [ lookup -d 0 cmlist $name ]
	set previousmatch [ lookup -d 0 pmlist $name ]
	set severity [ lookup -d "error" sevlist $name ]
	lappend patternList [ list $rowstatus $name $desc $pattern $kpattern $severity $currentmatch $previousmatch ] 
   }

   setAttribute patternList $patternList
}

proc setPattern { attrib name value } {
    ddl print debug "setPattern: $attrib $name $value\n"
    set key [ nameToKey $name ] 
   
    #
    # set newpattern to trigger pscan if new pattern is added
    # 
 
    define $attrib $key $value
     
    return $value 
}

proc addPattern { index name rowstatus } {

    ddl print debug "addPattern: $index $name $rowstatus"

    set key [ nameToKey $name ] 

    define rowstatus $key $rowstatus 
}


proc removePattern { index name } {
    #
    # clear any alarm status and attributes (limits, status command, and acks) 
    # associated with this row
    #
    set tableObject [ locate patternTable.patternEntry ]
    if { $tableObject != "" } {
        toe_send $tableObject cleanupRow $name CLEAR_PARMS
    } else {
        ddl print warning "could not find patternEntry object while removing entry $name in module [ getModuleParam moduleSpec ] \n"
    }
    #
    # clear pattern and associated data
    #
    undefine rowstatus $name 
    undefine plist $name 
    undefine dlist $name
    undefine kplist $name
    undefine cmlist $name 
    undefine pmlist $name
    undefine sevlist $name


}
#
# refreshPatternTable the pattern from the persistent data, if
# the module is initialy loaded
#
proc refreshPatternTable {} {
	ddl print debug "refreshPatternTable..."
	if { [ entries plist ] == "" } {
		getPatternList
	}
	set result ""
	
 	sliceforeach name pattern plist {
 		set desc [ lookup -d "" dlist $name ]
		set currentmatch [ lookup -d 0 cmlist $name ]
		set previousmatch [ lookup -d 0 pmlist $name ]
		set kpattern [ lookup -d "" kplist $name ]
 		set currentmatchDesc [ sec2date $currentmatch ]
		set previousmatchDesc [ sec2date $previousmatch ]
		set delta [ getDelta $currentmatch $previousmatch ]
		set severity [ lookup -d "error" sevlist $name ]  
 		lappend result [ list $name $desc $pattern $kpattern $severity $currentmatch $currentmatchDesc $previousmatch $previousmatchDesc $delta ]
  
	}
		 
	return $result
}


proc getPatternList {} {

	# get the list from the disk
 	
	set PatternList [ getAttribute patternList "" ]
	undefine rowstatus
	undefine plist
	undefine dlist
	undefine kplist
	undefine cmlist
	undefine pmlist
	undefine sevlist
	
	
	# undefine anything we might already have

	
	
	if { $PatternList != "" } {
		set idx 1
		set rowstatusid [ locate *patternEntry.rowstatus ]
		foreach entry $PatternList {
			lextract $entry 0 rowstatus 1 name 2 desc 3 pattern 4 kpattern 5 severity 6 currentmatch 7 previousmatch 
			define rowstatuslist $name $rowstatus
			define plist $name $pattern
			define kplist $name $kpattern
			define dlist $name $desc
			define cmlist $name $currentmatch
			define pmlist $name $previousmatch
			define sevlist $name $severity
			
			toe_send $rowstatusid setRowStatus $idx $name $rowstatus
			incr idx
		}		
 
  	}
}
#
# Read the stored pattern and pattern name data
#
proc receivePatternList {} {
	set result ""	
	set object [ locate panicpattern ]
	toe_begin $object
	 
		sliceforeach name pattern plist {
			lappend result $pattern
		}
	toe_end
	
	return $result
}

proc receivePatternNameList {} {
	set result ""	
	set object [ locate panicpattern ]
	toe_begin $object
	 
		sliceforeach name pattern plist {
			lappend result $name
		}
	toe_end
	
	return $result
}

proc receiveKeyPatternList {} {
	set result ""	
	set object [ locate panicpattern ]
	toe_begin $object
	 
		sliceforeach name kpattern kplist {
			lappend result $kpattern
		}
	toe_end
	
	return $result
}

#
# Function transform seconds into the form 0d 0h 0m 0s
#
proc sec2day { seconds } {

	set days 0
	set hours 0
	set mins 0
	set secs 0
	set remain 0
 	set days [ expr int($seconds/86400) ]
	set remain [ expr $seconds%86400 ]
 	if { $remain != 0 } {
 		set hours [ expr int($remain/3600) ]
		set remain [ expr $remain%3600 ]
 	        if { $remain != 0 } {
 			set mins [ expr int($remain/60) ]
			set remain [ expr $remain%60 ]
 			if { $remain != 0 } {
				set secs $remain
			}
		}
	}

	return [ format "%dd %dh %dm %ds" $days $hours $mins $secs ]
}	


#
# refreshCPUTable reads the initial data from the persistant storage.
#
proc refreshCPUTable {} {
	ddl print debug "refreshCPUTable..."
	
	set idtable [ locate *cpupanicEntry.Id ]
	set cputbl [ locate *cpupanicEntry ]
	set rowstatusid [ locate *cpupanicEntry.rowstatus ]
	
	
	set result ""
	if { [ toe_send $cputbl entries indexByName ] == "" } {
		set cpuPanicList [ getAttribute cpuPanicList "" ]
		set idx 1
		foreach entry $cpuPanicList {
			lextract $entry 0 rowstatus 1 id 2 cpu 3 previousp 4 currentp 5 newpanic 6 type

			set delta [ getDelta $currentp $previousp ]
			set cpd [ sec2date $currentp ]
			set ppd [ sec2date $previousp ]
			lappend result [ list $id $cpu $ppd $previousp $cpd $currentp 0 $delta $type ]
			toe_send $rowstatusid setRowStatus $idx $id $rowstatus
			incr idx
		}	
	} else {

		toe_begin $cputbl
     		sliceforeach index value indexByName {
     			set entry [ getRowData $index ]
    			lextract $entry 1 rowdata
    			lextract $rowdata 0 rowstatus 1 id 2 cpu 4 previousp 6 currentp 7 newpanic 9 type

			set delta [ getDelta $currentp $previousp ]
 			set cpd [ sec2date $currentp ]
			set ppd [ sec2date $previousp ]
    			lappend result [ list $id $cpu $ppd $previousp $cpd $currentp $newpanic $delta $type ]
     			toe_send $rowstatusid setRowStatus $value $id $rowstatus
    		}
    		toe_end		
	}

	
	return $result
}

proc getDelta { curr prev } {
	set result ""
	if { $prev > 0 } {
		set result [ sec2day [ expr $curr-$prev ] ]
	}
	
	return $result 

}
#
# Write Panic to CPU list to the persistant storage
#
proc syncCPUList {} {
	
	set idtable [ locate *cpupanicEntry.Id ]
	set cputbl [ locate *cpupanicEntry ]
	set cpuPanicList ""
	toe_begin $idtable
	sliceforeach index value data {
    		set entry [ toe_send $cputbl getRowData $index ]
    		lextract $entry 1 rowdata
		lextract $rowdata 0 rowstatus 1 id 2 cpu 4 previousp 6 currentp 7 newpanic 9 type
		lappend cpuPanicList [ list $rowstatus $id $cpu $previousp $currentp 0 $type ] 
   	}
   	toe_end

   	setAttribute cpuPanicList $cpuPanicList 
}

# This is a little bit of a hack
# I used the row disable function to delete all row with the same CPU ID
# from the Panic per CPU table.
# I have to investigate to find an other solution

proc resetCPUList { index rowname } {

	ddl print debug "resetCPUList.$index $rowname.."
	set rowstatusid [ locate *cpupanicEntry.rowstatus ]
	set cputbl [ locate *cpupanicEntry ]
	set object_cp [ locate cpupanicTable ]
	
	# Extract the CPU Number to be reseted
	set entry [ toe_send $cputbl getRowData $rowname ]
	lextract $entry 1 rowdata
	lextract $rowdata 2 cpuID
	 
	# Search for all entries with the selected ID and delete them.
	toe_begin $cputbl
     		sliceforeach rname idx indexByName {
     				ddl print info "indexByName $idx $rname"
    				set entry [ getRowData $rname ]
    				lextract $entry 1 rowdata
    				lextract $rowdata 2 cpu
 				if { $cpu == $cpuID } {
					ddl print info "setRowStatus $idx $rname 6"
     					toe_send $rowstatusid setRowStatus $idx $rname 6
					cleanupRow $idx CLEAR_PARMS
				}
    		}
    	toe_end		
	
	# Refresh and sync persistent data 
	toe_send $object_cp syncCPUList
	toe_send $cputbl setTrapInfo refreshOID
	
}

proc removeCPUEntry { index rowname } {
	
	
	set cputbl [ locate cpupanicTable.cpupanicEntry ]
	set object_cp [ locate cpupanicTable ]	
	
	toe_send $cputbl cleanupRow $rowname CLEAR_PARMS
	
	toe_send $object_cp syncCPUList
	toe_send $cputbl setTrapInfo refreshOID
	
}

proc getCPUList {} {

}
 
# Tests for the exsistance of a file <filename>
# and returns the modification time in human readable
# format

proc getMTime { filename } {
	set result ""
	if { [ file exists $filename ] } {
		set mTime [ file mtime $filename ]
		set result [ list [ clock format $mTime -format {%D %T} ] ]
		
	} else {
		set result [ list "Can't read $filename" ]
	}
	return $result
}

#
# Transforms seconds into a date in the format DD/MM/YY HH:MM:SS
#
proc sec2date { date } {
	set result ""

	if { $date > 0 } {
		set result [ clock format $date -format {%D %T} -gmt true ]
	}  
	return $result
}
